from __future__ import annotations
from typing import TYPE_CHECKING
from typing import Any

from fenix.base.exchange import Exchange

from fenix.base.constants import Side
from fenix.base.constants import ExchangeCode
from fenix.base.constants import Product
from fenix.base.constants import Validity
from fenix.base.constants import Variety
from fenix.base.constants import Status
from fenix.base.constants import Order
from fenix.base.constants import Root
from fenix.base.constants import WeeklyExpiry
from fenix.base.constants import UniqueID


from fenix.base.errors import InputError
from fenix.base.errors import TokenDownloadError

if TYPE_CHECKING:
    from requests.models import Response


class kotak(Exchange):
    """
    Kotak fenix Broker Class.

    Returns:
        fenix.kotak: fenix Kotak Broker Object.
    """


    # Market Data Dictonaries

    indices = {}
    eq_tokens = {}
    nfo_tokens = {}
    token_params = ["user_id", "password", "consumer_key", "access_token"]
    id = 'kotak'
    _session = Exchange._create_session()


    # Base URLs

    base_urls = {
        "api_doc": "https://ctradeapi.kotaksecurities.com/devportal/apis",
        "git_sdk": "https://github.com/paramatrixtech/ksapi",
        "base": "https://tradeapi.kotaksecurities.com/apim",
        "market_data": "https://preferred.kotaksecurities.com/security/production",
    }

    # Access Token Generation URLs

    token_urls = {
        "login": f"{base_urls['base']}/session/1.0/session/login/userid",
        "session_token": f"{base_urls['base']}/session/1.0/session/2FA/oneTimeToken",
    }


    urls = {
        "place_order": f"{base_urls['base']}/placeOrder/executePlaceOrder",
        "modify_order": f"{base_urls['base']}/placeOrder/modifyOrder",
        "cancel_order": f"{base_urls['base']}/placeOrder/cancelOrder",
        "exit_bracket_order": f"{base_urls['base']}/placeOrder/exitBracketOrder",
        "orderbook": f"{base_urls['base']}/placeOrder/fetchOrderBook",
        "tradebook": f"{base_urls['base']}/placeOrder/fetchTradeBook",
        "order_history": f"{base_urls['base']}/placeOrder/orderHistory",
        "positions": f"{base_urls['base']}/positionAndHoldings/positionBook",
        "holdings": f"{base_urls['base']}/positionAndHoldings/holdings",
        "sqoff_position": f"{base_urls['base']}/positionAndHoldings/sqrOofPosition",
        "rms_limits": f"{base_urls['base']}/limits/getRmsLimits",
        "profile": f"{base_urls['base']}/customer/accountDetails",
    }

    order_cancel_url = f"{base_urls['base']}/orders/1.0/orders"
    fetch_order_url = f"{base_urls['base']}/reports/1.0/orders"
    fetch_trade_book_url = f"{base_urls['base']}/reports/1.0/trades"

    place_order_urls = {
        Product.NRML: f"{base_urls['base']}/orders/1.0/order/normal",
        Product.SM: f"{base_urls['base']}/orders/1.0/order/supermultiple",
        "SOR": f"{base_urls['base']}/orders/1.0/order/sor",
        "MTF": f"{base_urls['base']}/orders/1.0/order/mtf",
        Product.MIS: f"{base_urls['base']}/orders/1.0/order/mis"
    }


    # Request Parameters Dictionaries

    req_side = {
        Side.BUY: "BUY",
        Side.SELL: "SELL"
    }

    req_product = {
        Product.MIS: "MIS",
        Product.NRML: "NORMAL",
        "SUPERMULTIPLE": "SUPERMULTIPLE",
        "SUPERMULTIPLEOPTION": "SUPERMULTIPLEOPTION",
        "MTF": "MTF",
    }

    req_validity = {
        Validity.DAY: "GFD",
        Validity.IOC: "IOC",
    }

    req_variety = {
        Variety.REGULAR: "REGULAR",
        Variety.AMO: "AMO",
        "SQUAREOFF": "SQUAREOFF"
    }


    # Response Parameters Dictionaries

    resp_validity = {
        "Good For Day": "DAY"
    }

    resp_product = {
        "NORMAL": Product.NRML
    }

    resp_status = {
        "OPN": Status.OPEN,
        "Traded": Status.FILLED,
        "NEWF": Status.PENDING,
    }


    # NFO Script Fetch


    @classmethod
    def data_datetime(cls):
        """
        Datetime object converted to a string format to be used in creating a market data URL.

        Returns:
            datetime string (str): Datetime string.
        """
        todaysdate = cls.current_datetime()
        weekday = todaysdate.weekday()
        days = 0
        if weekday >= 5:
            days = 3

        date_obj = cls.time_delta(todaysdate, days, dtformat="%d_%m_%Y")

        return date_obj

    @classmethod
    def create_eq_tokens(cls) -> dict:
        """
        Gives Indices Info for F&O Segment.
        Stores them in the Kotak.indices Dictionary.

        Returns:
            dict: Unified fenix indices format.
        """
        date_obj = cls.data_datetime()
        link = f"{cls.base_urls['market_data']}/TradeApiInstruments_Cash_{date_obj}.txt"

        df = cls.data_reader(link, filetype='csv', sep="|")

        df = df[df['instrumentType'] == "EQ"]
        df = df[["exchange", "instrumentName", "instrumentToken", "tickSize", "lotSize"]]


        df.rename({"instrumentToken": "Token", "instrumentName": "Symbol",
                   'tickSize': 'TickSize', "lotSize": "LotSize", "exchange": "Exchange"
                   }, axis=1, inplace=True)

        df_bse = df[df['Exchange'] == ExchangeCode.BSE]
        df_bse.drop_duplicates(subset=['Symbol'], keep='first', inplace=True)
        df_bse.set_index(df_bse['Symbol'], inplace=True)


        df_nse = df[df['Exchange'] == ExchangeCode.NSE]
        df_nse.set_index(df_nse['Symbol'], inplace=True)

        cls.eq_tokens[ExchangeCode.NSE] = df_nse.to_dict(orient='index')
        cls.eq_tokens[ExchangeCode.BSE] = df_bse.to_dict(orient='index')

        return cls.eq_tokens


    @classmethod
    def create_indices(cls) -> dict:
        """
        Gives Indices Info for F&O Segment.
        Stores them in the kotak.indices Dictionary.

        Returns:
            dict: Unified fenix indices format.
        """
        date_obj = cls.data_datetime()
        link = f"{cls.base_urls['market_data']}/TradeApiInstruments_Cash_{date_obj}.txt"

        df = cls.data_reader(link, filetype='csv', sep="|")
        df = df[df['instrumentType'] == "IN"][["instrumentName", "instrumentToken"]]

        df.rename({"instrumentName": "Symbol", "instrumentToken": "Token"}, axis=1, inplace=True)
        df.index = df['Symbol']

        indices = df.to_dict(orient='index')

        indices[Root.BNF] = indices["NIFTY BANK"]
        indices[Root.NF] = indices["NIFTY 50"]
        indices[Root.FNF] = indices["FINNIFTY"]
        indices[Root.MIDCPNF] = indices["NIFTY MIDCAP 50"]  # could not find NIFTY MIDCAP SELECT

        cls.indices = indices

        return indices

    @classmethod
    def create_nfo_tokens(cls) -> dict:
        """
        Creates BANKNIFTY & NIFTY Current, Next and Far Expiries;
        Stores them in the kotak.nfo_tokens Dictionary.

        Raises:
            TokenDownloadError: Any Error Occured is raised through this Error Type.
        """
        try:
            date_obj = cls.data_datetime()
            link = f"{cls.base_urls['market_data']}/TradeApiInstruments_FNO_{date_obj}.txt"

            df = cls.data_reader(link, filetype='csv', sep="|")

            df = df[
                (
                    (df['instrumentName'] == 'BANKNIFTY') |
                    (df['instrumentName'] == 'NIFTY') |
                    (df['instrumentName'] == 'FINNIFTY') |
                    (df['instrumentName'] == 'MIDCPNIFTY')
                ) &
                (
                    (df['optionType'] != "XX")
                )
            ]

            df.rename({"instrumentToken": "Token", "instrumentName": "Root", "expiry": "Expiry",
                       "optionType": "Option", 'tickSize': 'TickSize', "lotSize": "LotSize",
                       "lastPrice": "LastPrice", "strike": "StrikePrice", "exchange": "Exchange"
                       },
                      axis=1, inplace=True)

            df = df[['Token', 'Expiry', 'Option',
                     'StrikePrice', 'LotSize',
                     'Root', 'LastPrice', 'TickSize', "Exchange"
                     ]]

            # df["Exchange"] = ExchangeCode.NFO  # Is NSE in Script Master but not Needed amywhere .
            df['StrikePrice'] = df['StrikePrice'].astype(int)
            df['Expiry'] = cls.pd_datetime(df['Expiry']).dt.date.astype(str)

            expiry_data = cls.jsonify_expiry(data_frame=df)
            cls.nfo_tokens = expiry_data

            return expiry_data

        except Exception as exc:
            raise TokenDownloadError({"Error": exc.args}) from exc


    # Headers & Json Parsers


    @classmethod
    def create_headers(cls,
                       params: dict,
                       ) -> dict[str, str]:
        """
        Generate Headers used to access Endpoints in Kotak.

        Parameters:
            Params (dict) : A dictionary which should consist the following keys:
                user_id (str): User ID of the Account.
                password (str): Password of the Account.
                consumer_key (str): Consumer Key of the Account API.
                access_token (str): Access Token of the Account API.

        Returns:
            dict[str, str]: Kotak Headers.
        """
        for key in cls.token_params:
            if key not in params:
                raise KeyError(f"Please provide {key}")

        headers01 = {
            "consumerKey": params["consumer_key"],
            "ip": "0.0.0.0",
            "appId": "APP1",
            "Content-Type": "application/json",
            "User-Agent": "KSTradeApi-python/1.0.0/python",
            "Authorization": f"Bearer {params['access_token']}"
        }

        json_data01 = {
            "userid": params["user_id"],
            "password": params["password"]
        }

        response01 = cls.fetch(method="POST", url=cls.token_urls["login"],
                               json=json_data01, headers=headers01)

        info01 = cls._json_parser(response01)
        one_time_token = info01['Success']['oneTimeToken']

        headers02 = {
            "oneTimeToken": one_time_token,
            "consumerKey": params["consumer_key"],
            "ip": "0.0.0.0",
            "appId": "APP1",
            "Accept": "application/json",
            "Content-Type": "application/json",
            "User-Agent": "KSTradeApi-python/1.0.0/python",
            "Authorization": f"Bearer {params['access_token']}"
        }

        json_data02 = {"userid": params["user_id"]}

        response02 = cls.fetch(method="POST", url=cls.token_urls["session_token"],
                               json=json_data02, headers=headers02)

        info02 = cls._json_parser(response02)

        headers = {
            "headers": {
                "Accept": "application/json",
                "Content-Type": "application/json",
                "User-Agent": "KSTradeApi-python/1.0.0/python",
                "consumerKey": params["consumer_key"],
                "sessionToken": info02["success"]["sessionToken"],
                "Authorization": f"Bearer {params['access_token']}"
            }
        }

        cls._session = cls._create_session()

        return headers

    @classmethod
    def _json_parser(cls,
                     response: Response
                     ) -> dict[Any, Any] | list[dict[Any, Any]]:
        """
        Parses the Json Repsonse Obtained from Broker.

        Parameters:
            response (Response): Json Response Obtained from Broker.

        Raises:
            ResponseError: Raised if any error received from broker.

        Returns:
            dict: json response obtained from exchange.
        """
        json_response = cls.on_json_response(response)
        # print(json_response)
        return json_response

    @classmethod
    def _orderbook_json_parser(cls,
                               order: dict,
                               ) -> dict[Any, Any]:
        """
        Parse Orderbook Order Json Response.

        Parameters:
            order (dict): Orderbook Order Json Response from Broker.

        Returns:
            dict: Unified fenix Order Response.
        """
        parsed_order = {
            Order.ID: order["orderId"],
            Order.USERID: order["tag"],
            Order.TIMESTAMP: cls.datetime_strp(order["orderTimestamp"], "%b %d %Y %I:%M:%S:%f%p"),
            Order.SYMBOL: order["instrumentName"],
            Order.TOKEN: order["instrumentToken"],
            Order.SIDE: cls.req_side.get(order["transactionType"], order["transactionType"]),
            Order.TYPE: "",
            Order.AVGPRICE: order["price"],
            Order.PRICE: order["price"],
            Order.TRIGGERPRICE: order["triggerPrice"],
            Order.QUANTITY: order["orderQuantity"],
            Order.FILLEDQTY: order["filledQuantity"],
            Order.REMAININGQTY: order["pendingQuantity"],
            Order.CANCELLEDQTY: 0,
            Order.STATUS: cls.resp_status.get(order["statusInfo"], order["statusInfo"]),
            Order.REJECTREASON: "",
            Order.DISCLOSEDQUANTITY: order["disclosedQuantity"],
            Order.PRODUCT: cls.resp_product.get(order["product"], order["product"]),
            Order.EXCHANGE: order["exchange"],
            Order.SEGMENT: order["exchange"],
            Order.VALIDITY: "DAY",
            Order.VARIETY: "",
            Order.INFO: order,
        }

        return parsed_order

    @classmethod
    def _tradebookhistory_json_parser(cls,
                                      order: dict
                                      ) -> dict[Any, Any]:
        """
        Parse Orderbook Order Json Response.

        Parameters:
            order (dict): Orderbook Order Json Response from Broker.

        Returns:
            dict: Unified fenix Order Response.
        """
        parsed_order = {
            Order.ID: order['exchOrderId'],
            Order.USERID: order['message'],
            Order.TIMESTAMP: cls.datetime_strp(order["activityTimestamp"], '%b %d %Y %I:%M:%S:%f%p'),
            Order.SYMBOL: "",
            Order.TOKEN: 0,
            Order.SIDE: "",
            Order.TYPE: "",
            Order.AVGPRICE: 0.0,
            Order.PRICE: order['price'],
            Order.TRIGGERPRICE: order['triggerPrice'],
            Order.QUANTITY: order['orderQuantity'],
            Order.FILLEDQTY: order['filledQuantity'],
            Order.REMAININGQTY: order['orderQuantity'] - order['filledQuantity'],
            Order.CANCELLEDQTY: 0,
            Order.STATUS: cls.resp_status.get(order['status'], order['status']),
            Order.REJECTREASON: "",
            Order.DISCLOSEDQUANTITY: 0,
            Order.PRODUCT: "",
            Order.SEGMENT: "",
            Order.VALIDITY: cls.resp_validity.get(order['validity'], order['validity']),
            Order.VARIETY: "",
        }

        return parsed_order

    @classmethod
    def _create_order_parser(cls,
                             response: Response,
                             key_to_check: str,
                             headers: dict
                             ) -> dict[Any, Any]:
        """
        Parse Json Response Obtained from Broker After Placing Order to get order_id
        and fetching the json repsone for the said order_id.

        Parameters:
            response (Response): Json Repsonse Obtained from broker after Placing an Order.
            headers (dict): headers to send order request with.

        Returns:
            dict: Unified fenix Order Response.
        """
        info = cls._json_parser(response)

        if 'fault' in info:
            detail = info['fault']

            order = {
                detail.ID: "0",
                Order.USERID: "0",
                Order.TIMESTAMP: cls.current_datetime(),
                Order.SYMBOL: "",
                Order.TOKEN: 0,
                Order.SIDE: "",
                Order.TYPE: "",
                Order.AVGPRICE: 0.0,
                Order.PRICE: 0.0,
                Order.TRIGGERPRICE: 0.0,
                Order.QUANTITY: 0,
                Order.FILLEDQTY: 0,
                Order.REMAININGQTY: 0,
                Order.CANCELLEDQTY: 0,
                Order.STATUS: Status.REJECTED,
                Order.REJECTREASON: detail['message'],
                Order.DISCLOSEDQUANTITY: 0,
                Order.PRODUCT: "",
                Order.SEGMENT: "",
                Order.VALIDITY: "",
                Order.VARIETY: ""
            }

            return order

        else:
            order_id = info['Success']['NSE']['orderId']
            data = cls.fetch_order(order_id=order_id, headers=headers)

            return data

            # if  "BSE"  not in info['Success']:

            #     order = info['Success']['NSE']

            #     data = {
            #         "id": order['orderId'],
            #         "userOrderId": order['tag'],
            #         "price": order['orderId'],
            #         "qauntity": order['quantity'],
            #         'message': order['message'],
            #         "info": info
            #             }

            #     return data

            # else:
            #     orderNSE = info['Success']['NSE']
            #     orderBSE = info['Success']['BSE']

            #     data = {
            #         "NSE": {
            #             "id": orderNSE['orderId'],
            #             "userOrderId": orderNSE['tag'],
            #             "price": orderNSE['orderId'],
            #             "qauntity": orderNSE['quantity'],
            #             "message": orderNSE['message']
            #             },
            #         "BSE": {
            #             "id": orderBSE['orderId'],
            #             "userOrderId": orderBSE['tag'],
            #             "price": orderBSE['orderId'],
            #             "qauntity": orderBSE['quantity'],
            #             "message": orderBSE['message']
            #             },
            #         "info": info
            #             }
            #     return data

        #
        # Sample Response
        #
        # {"Success":{
        #     "NSE":{
        #         "message":"Your Order has been Placed and Forwarded to the Exchange: 2230131034372.",
        #         "orderId":2230131034372,
        #         "price":0,
        #         "quantity":1,
        #         "tag":"HD"
        #         }
        #     }
        #  }
        #
        #


    # Order Functions


    def create_order(cls, token: int, side: str, price: float, triggerprice: float, uniqueid: int, quantity: int,
                     variety: str, product: str, validity: str, headers: dict, orderType: str = "", exchange: str = ""):

        """
        create a trade order

        :param int token: token of the script of the market to create an order in
        :param str side: 'BUY' or 'SELL'
        :param float price: price of the order
        :param float triggerprice: 0 for market order, non-zero for limit
        :param str uniqueid: a unique id to denote your order
        :param int quantity: quantity of shares
        :param str ordertype: Not Needed for this Exchange
        :param str exchange: Not Needed for this Exchange
        :param str product: 'NRML', 'SM', 'SOR', 'MTF', 'MIS'
        :param str validity: 'DAY' or 'IOC'
        :param str variety: "REGULAR" or "AMO"
        :returns json: a json response of order
        """

        json_data = {
            'instrumentToken': token,
            'transactionType': side,
            'quantity': quantity,
            'price': price,
            'validity': cls.key_mapper(cls.req_validity, validity, "validity"),
            'variety': cls.key_mapper(cls.req_variety, variety, "variety"),
            'disclosedQuantity': 0,
            'triggerPrice': triggerprice,
            'tag': uniqueid
        }

        url = cls.key_mapper(cls.place_order_urls, product, "product")

        response = cls.fetch(method="POST", url=url, data=json_data, headers=headers["headers"])

        return cls._create_order_parser(response=response, headers=headers)

    def market_order(cls, token: int, side: str, uniqueid: int, quantity: int, headers: dict,
                     product: str = Product.MIS, validity: str = Validity.DAY, variety: str = Variety.REGULAR, exchange: str = ""):

        """
        create a market order
        :param int token: token of the script of the market to create an order in
        :param str side: 'BUY' or 'SELL'
        :param str uniqueid: a unique id to denote your order
        :param int quantity: quantity of shares
        :param str exchange: Not Needed for this Exchange
        :param str product: 'NRML', 'SM', 'SOR', 'MTF', 'MIS'
        :param str validity: 'DAY' or 'IOC'
        :param str variety: "REGULAR" or "AMO"
        :returns json: a json response of order
        """

        json_data = {
            'instrumentToken': token,
            'transactionType': side,
            'quantity': quantity,
            'price': 0,
            'validity': cls.key_mapper(cls.req_validity, validity, "validity"),
            'variety': cls.key_mapper(cls.req_variety, variety, "variety"),
            'disclosedQuantity': 0,
            'triggerPrice': 0,
            'tag': uniqueid
        }

        final_url = cls.key_mapper(cls.order_place_urls, product, "product")

        response = cls.fetch(method="POST", url=final_url, json=json_data, headers=headers["headers"])

        return cls._create_order_parser(response=response, headers=headers)

    def limit_order(cls, price: float, quantity: int,
                    token: int, side: str, uniqueid: int, headers: dict,
                    product: str = Product.MIS, validity: str = Validity.DAY, variety: str = Variety.REGULAR, exchange: str = ""):

        """
        create a market order
        :param int token: token of the script of the market to create an order in
        :param str side: 'BUY' or 'SELL'
        :param str uniqueid: a unique id to denote your order
        :param int quantity: quantity of shares
        :param str exchange: Not Needed for this Exchange
        :param str product: 'NRML', 'SM', 'SOR', 'MTF', 'MIS'
        :param str validity: 'DAY' or 'IOC'
        :param str variety: "REGULAR" or "AMO"
        :returns json: a json response of order
        """

        json_data = {
            'instrumentToken': token,
            'transactionType': side,
            'quantity': quantity,
            'price': price,
            'validity': cls.key_mapper(cls.req_validity, validity, "validity"),
            'variety': cls.key_mapper(cls.req_variety, variety, "variety"),
            'disclosedQuantity': 0,
            'triggerPrice': 0,
            'tag': uniqueid
        }

        final_url = cls.key_mapper(cls.order_place_urls, product, "product")

        response = cls.fetch(method="POST", url=final_url, json=json_data, headers=headers["headers"])

        return cls._create_order_parser(response=response, headers=headers)

    def sl_order(cls, price: float, triggerprice: float, quantity: int,
                 token: int, side: str, uniqueid: int, headers: dict,
                 product: str = Product.MIS, validity: str = Validity.DAY, variety: str = Variety.REGULAR, exchange: str = ""):

        """
        create a market order
        :param int token: token of the script of the market to create an order in
        :param str side: 'BUY' or 'SELL'
        :param str uniqueid: a unique id to denote your order
        :param int quantity: quantity of shares
        :param str exchange: Not Needed for this Exchange
        :param str product: 'NRML', 'SM', 'SOR', 'MTF', 'MIS'
        :param str validity: 'DAY' or 'IOC'
        :param str variety: "REGULAR" or "AMO"
        :returns json: a json response of order
        """

        json_data = {
            'instrumentToken': token,
            'transactionType': side,
            'quantity': quantity,
            'price': price,
            'validity': cls.key_mapper(cls.req_validity, validity, "validity"),
            'variety': cls.key_mapper(cls.req_variety, variety, "variety"),
            'disclosedQuantity': 0,
            'triggerPrice': triggerprice,
            'tag': uniqueid
        }

        final_url = cls.key_mapper(cls.order_place_urls, product, "product")

        response = cls.fetch(method="POST", url=final_url, json=json_data, headers=headers["headers"])

        return cls._create_order_parser(response=response, headers=headers)


    # NFO Order Functions


    @classmethod
    def market_order_nfo(cls, option: str, strikeprice: int,
                         side: str, quantity: int,
                         headers: dict,
                         root: str = Root.BNF,
                         expiry: str = WeeklyExpiry.CURRENT,
                         uniqueid: str = UniqueID.MARKETORDER,
                         product: str = Product.MIS, variety: str = "REGULAR",
                         validity: str = Validity.DAY):

        """
        create a market order
        :param str option: 'CE' or 'PE'
        :param str root: 'BANKNIFTY' or 'NIFTY'
        :param int strikeprice: Strike Price of the Option
        :param str side: 'BUY' or 'SELL'
        :param str expiry: 'CURRENT', 'NEXT', 'FAR'
        :param str uniqueid: a unique id to denote your order
        :param int quantity: quantity of shares
        :param str product: 'NRML', 'MIS', 'SM' (Super Multiple Order), 'SOR' (Smart Order Routing), 'MTF' (Margin Tading Facility)
        :param str validity: 'DAY' or 'IOC'
        :param str variety: "REGULAR" or "AMO"
        :returns json: a json response of order
        """

        if not cls.nfo_tokens:
            cls.create_nfo_tokens()

        detail = cls.nfo_tokens[expiry][root][option]
        detail = detail.get(strikeprice, None)

        if not detail:
            raise KeyError(f"StrikePrice: {strikeprice} Does not Exist")

        token = detail['Token']

        json_data = {
            'instrumentToken': token,
            'transactionType': side,
            'quantity': quantity,
            'price': 10,
            'validity': cls._key_mapper(cls.req_validity, validity, "validity"),
            'variety': cls._key_mapper(cls.req_variety, variety, "variety"),
            'disclosedQuantity': 0,
            'triggerPrice': 0,
            'tag': uniqueid
        }

        final_url = cls._key_mapper(cls.place_order_urls, product, "product")

        response = cls.fetch(method="POST", url=final_url, json=json_data, headers=headers["headers"])

        return cls._create_order_parser(response, headers)

    def limit_order_nfo(cls, price: float, quantity: int,
                        option: str, root: str, strikeprice: int, side: str, headers: dict,
                        expiry: str = WeeklyExpiry.CURRENT, uniqueid: str = UniqueID.LIMITORDER,
                        product: str = Product.MIS, variety: str = "REGULAR", validity: str = Validity.DAY):

        """
        create a market order
        :param str option: 'CE' or 'PE'
        :param str root: 'BANKNIFTY' or 'NIFTY'
        :param int strikeprice: Strike Price of the Option
        :param str side: 'BUY' or 'SELL'
        :param str expiry: 'CURRENT', 'NEXT', 'FAR'
        :param str uniqueid: a unique id to denote your order
        :param int quantity: quantity of shares
        :param str product: 'NRML', 'MIS', 'SM' (Super Multiple Order), 'SOR' (Smart Order Routing), 'MTF' (Margin Tading Facility)
        :param str validity: 'DAY' or 'IOC'
        :param str variety: "REGULAR" or "AMO"
        :returns json: a json response of order
        """

        if not cls.Global:
            cls.expiry_markets()

        detail = cls.Global[expiry][root][option]
        detail = detail.get(strikeprice, None)

        if not detail:
            raise KeyError(f"StrikePrice: {strikeprice} Does not Exist")

        token = detail['Token']

        json_data = {
            'instrumentToken': token,
            'transactionType': side,
            'quantity': quantity,
            'price': price,
            'validity': cls.key_mapper(cls.req_validity, validity, "validity"),
            'variety': cls.key_mapper(cls.req_variety, variety, "variety"),
            'disclosedQuantity': 0,
            'triggerPrice': 0.0,
            'tag': uniqueid
        }

        final_url = cls.key_mapper(cls.place_order_urls, product, "product")

        response = cls.fetch(method="POST", url=final_url, json=json_data, headers=headers["headers"])

        return cls._create_order_parser(response=response, headers=headers)

    def sl_order_nfo(cls, price: float, triggerprice: float,
                     option: str, root: str, strikeprice: int, side: str, quantity: int, headers: dict,
                     expiry: str = WeeklyExpiry.CURRENT, uniqueid: str = UniqueID.SLORDER,
                     product: str = Product.MIS, variety: str = "REGULAR", validity: str = Validity.DAY):

        """
        create a market order
        :param str option: 'CE' or 'PE'
        :param str root: 'BANKNIFTY' or 'NIFTY'
        :param int strikeprice: Strike Price of the Option
        :param str side: 'BUY' or 'SELL'
        :param str expiry: 'CURRENT', 'NEXT', 'FAR'
        :param str uniqueid: a unique id to denote your order
        :param int quantity: quantity of shares
        :param str product: 'NRML', 'MIS', 'SM' (Super Multiple Order), 'SOR' (Smart Order Routing), 'MTF' (Margin Tading Facility)
        :param str validity: 'DAY' or 'IOC'
        :param str variety: "REGULAR" or "AMO"
        :returns json: a json response of order
        """

        if not cls.Global:
            cls.expiry_markets()

        detail = cls.Global[expiry][root][option]
        detail = detail.get(strikeprice, None)

        if not detail:
            raise KeyError(f"StrikePrice: {strikeprice} Does not Exist")

        token = detail['Token']

        json_data = {
            'instrumentToken': token,
            'transactionType': side,
            'quantity': quantity,
            'price': price,
            'validity': cls.key_mapper(cls.req_validity, validity, "validity"),
            'variety': cls.key_mapper(cls.req_variety, variety, "variety"),
            'disclosedQuantity': 0,
            'triggerPrice': triggerprice,
            'tag': uniqueid
        }

        final_url = cls.key_mapper(cls.place_order_urls, product, "product")

        response = cls.fetch(method="POST", url=final_url, json=json_data, headers=headers["headers"])

        return cls._create_order_parser(response=response, headers=headers)


    # Order Details, OrderBook & TradeBook


    @classmethod
    def fetch_raw_orderbook(cls,
                            headers: dict
                            ) -> list[dict]:
        """
        Fetch Raw Orderbook Details, without any Standardaization.

        Parameters:
            headers (dict): headers to send fetch_orders request with.

        Returns:
            list[dict]: Raw Broker Orderbook Response.
        """
        response = cls.fetch(method="GET", url=cls.fetch_order_url, headers=headers["headers"])
        return cls._json_parser(response)

    @classmethod
    def fetch_orderbook(cls,
                        headers: dict
                        ) -> list[dict]:
        """
        Fetch Orderbook Details.

        Parameters:
            headers (dict): headers to send fetch_orders request with.

        Returns:
            list[dict]: List of dicitonaries of orders using fenix Unified Order Response.
        """
        info = cls.fetch_raw_orderbook(headers=headers)

        orders = []
        for order in info['success']:
            detail = cls._orderbook_json_parser(order)
            orders.append(detail)

        return orders

    @classmethod
    def fetch_tradebook(cls,
                        headers: dict
                        ) -> list[dict]:
        """
        Fetch Tradebook Details.

        Parameters:
            headers (dict): headers to send fetch_orders request with.

        Returns:
            list[dict]: List of dicitonaries of orders using fenix Unified Order Response.
        """
        response = cls.fetch(method="GET", url=cls.fetch_trade_book_url, headers=headers["headers"])
        info = cls._json_parser(response)

        orders = []
        for order in info['success']:
            detail = cls._orderbook_json_parser(order)

            orders.append(detail)

        return orders

    @classmethod
    def fetch_orders(cls,
                     headers: dict
                     ) -> list[dict]:
        """
        Fetch OrderBook Details which is unified across all brokers.
        Use This if you want Avg price, etc. values which sometimes unavailable
        thorugh fetch_orderbook.

        Paramters:
            order_id (str): id of the order.

        Raises:
            InputError: If order does not exist.

        Returns:
            dict: fenix Unified Order Response.
        """
        return cls.fetch_orderbook(headers=headers)

    @classmethod
    def fetch_order(cls,
                    order_id: str,
                    headers: dict
                    ) -> dict[Any, Any]:
        """
        Fetch Order Details.

        Paramters:
            order_id (str): id of the order.

        Raises:
            InputError: If order does not exist.

        Returns:
            dict: fenix Unified Order Response.
        """
        order_id = str(order_id)
        info = cls.fetch_raw_orderbook(headers=headers)

        for order in info['success']:
            if order["orderId"] == order_id:
                detail = cls._orderbook_json_parser(order)
                return detail

        raise InputError({"This orderid does not exist."})

    @classmethod
    def fetch_tradebook_order(cls,
                              order_id: str,
                              headers: dict
                              ) -> dict[Any, Any]:
        """
        Fetch  TradeBook Order Details.

        Paramters:
            order_id (str): id of the order.

        Raises:
            InputError: If order does not exist.

        Returns:
            dict: fenix Unified Order Response.
        """
        final_url = f"{cls.fetch_trade_book_url}/{order_id}"

        response = cls.fetch(method="GET", url=final_url, headers=headers["headers"])
        info = cls._json_parser(response)

        detail = info['success'][-1]
        order = cls._tradebookhistory_json_parser(detail)

        return order


    # Order Modification & Sq Off


    @classmethod
    def cancel_order(cls,
                     order_id: str,
                     headers: dict
                     ) -> dict[Any, Any]:
        """
        Cancel an open order.

        Parameters:
            order_id (str): id of the order.
            headers (dict): headers to send cancel_order request with.

        Returns:
            dict: fenix Unified Order Response.
        """
        final_url = f"{cls.order_cancel_url}/{order_id}"

        response = cls.fetch(method="DELETE", url=final_url, headers=headers["headers"])

        return cls._create_order_parser(response=response, headers=headers)



__all__ = [
    "Side",
    "Root",
    "WeeklyExpiry",
    "OrderType",
    "ExchangeCode",
    "Product",
    "Validity",
    "Variety",
    "Status",
    "Order",
    "Position",
    "Profile",
    "UniqueID",
]


class Side:
    """
    Order Side Constants.
    """
    BUY = "BUY"
    SELL = "SELL"


class Root:
    """
    BANKNIFTY & NIFTY Constants.
    """
    BNF = "BANKNIFTY"
    NF = "NIFTY"
    FNF = "FINNIFTY"
    MIDCPNF = "MIDCPNIFTY"


class WeeklyExpiry:
    """
    Weekly Expiry Constants.
    """
    CURRENT = "CURRENT"
    NEXT = "NEXT"
    FAR = "FAR"
    EXPIRY = "Expiry"
    LOTSIZE = "LotSize"


class OrderType:
    """
    Order Type Constants.
    """
    MARKET = "MARKET"
    LIMIT = "LIMIT"
    SLM = "SLM"
    SL = "SL"


class ExchangeCode:
    """
    Exchange Code Constants.
    """
    NSE = "NSE"  # NSE Equity
    NFO = "NFO"  # NSE F&O
    BSE = "BSE"  # BSE Equity
    BFO = "BFO"  # BSE F&O
    NCO = "NCO"  # NSE Commodities
    BCO = "BCO"  # BSE Commodities
    BCD = "BCD"  #
    MCX = "MCX"  # Multi Commodity Exchange F&O
    CDS = "CDS"  #


class Product:
    """
    Product Type Constants.
    """
    CNC = "CNC"
    NRML = "NRML"
    MARGIN = "MARGIN"
    MIS = "MIS"
    BO = "BO"
    CO = "CO"
    SM = "SM"  # SuperMutilple


class Validity:
    """
    Order Validity Constnats.
    """
    DAY = "DAY"
    IOC = "IOC"
    GTD = "GTD"
    GTC = "GTC"
    FOK = "FOK"
    TTL = "TTL"


class Variety:
    """
    Order Variety Constants.
    """
    REGULAR = "REGULAR"
    STOPLOSS = "STOPLOSS"
    AMO = "AMO"
    BO = "BO"
    CO = "CO"
    ICEBERG = "ICEBERG"
    AUCTION = "AUCTION"


class Status:
    """
    Order Status Constants.
    """
    PENDING = "PENDING"
    OPEN = "OPEN"
    PARTIALLYFILLED = "PARTIALLYFILLED"
    FILLED = "FILLED"
    REJECTED = "REJECTED"
    CANCELLED = "CANCELLED"
    MODIFIED = "MODIFIED"


class Order:
    """
    Unified Order Response Dicitonary Keys Constants.
    """
    ID = "id"
    USERID = "userOrderId"
    CLIENTID = "clientId"
    TIMESTAMP = 'timestamp'
    SYMBOL = "symbol"
    TOKEN = "token"
    SIDE = "side"
    TYPE = "type"
    AVGPRICE = "avgPrice"
    PRICE = "price"
    TRIGGERPRICE = "triggerPrice"
    TARGETPRICE = "targetPrice"
    STOPLOSSPRICE = "stoplossPrice"
    TRAILINGSTOPLOSS = "trailingStoploss"
    QUANTITY = "quantity"
    FILLEDQTY = "filled"
    REMAININGQTY = "remaining"
    CANCELLEDQTY = "cancelleldQty"
    STATUS = "status"
    REJECTREASON = "rejectReason"
    DISCLOSEDQUANTITY = "disclosedQuantity"
    PRODUCT = "product"
    SEGMENT = "segment"
    EXCHANGE = "exchange"
    VALIDITY = "validity"
    VARIETY = "variety"
    INFO = "info"


class Position:
    """
    Unified Account Positions Response Dicitonary Keys Constants.
    """
    SYMBOL = "symbol"
    TOKEN = "token"
    NETQTY = "netQty"
    AVGPRICE = "avgPrice"
    MTM = "mtm"
    PNL = "pnl"
    BUYQTY = "buyQty"
    BUYPRICE = "buyPrice"
    SELLQTY = "sellQty"
    SELLPRICE = "sellPrice"
    LTP = "ltp"
    PRODUCT = "product"
    EXCHANGE = "exchange"
    INFO = "info"


class Profile:
    """
    Unified Account Profile Response Dicitonary Keys Constants.
    """
    CLIENTID = "clientId"
    NAME = "name"
    EMAILID = "emailId"
    MOBILENO = "mobileNo"
    PAN = "pan"
    ADDRESS = "address"
    BANKNAME = "bankName"
    BANKBRANCHNAME = "bankBranchName"
    BANKACCNO = "bankAccNo"
    EXHCNAGESENABLED = "exchangesEnabled"
    ENABLED = "enabled"
    INFO = "info"


class UniqueID:
    """
    Default Unique Order ID Constants.
    """
    DEFORDER= "KronosOrder"
    MARKETORDER = "MarketOrder"
    LIMITORDER = "LIMITOrder"
    SLORDER = "SLOrder"
    SLMORDER = "SLMOrder"
    DEFORDERNO = 100
    MARKETORDERNO = 101
    LIMITORDERNO = 102
    SLORDERNO = 102
    SLMORDERNO = 103



from fenix.symphony import symphony


class kunjee(symphony):
    """
    Kunjee fenix Broker Class

    Returns:
        fenix.kunjee: fenix Kunjee Broker Object
    """


    # Market Data Dictonaries

    id = 'kunjee'


    # Base URLs

    base_urls = {
        "api_documentation_url": "https://trade.kunjee.net:3000/doc/interactive",
        "market_data_url": "https://trade.kunjee.net:3000/apimarketdata/instruments/master",
        "base_url": "https://trade.kunjee.net:3000/interactive",
        "access_token": "https://trade.kunjee.net:3000/interactive/user/session"
    }



from __future__ import annotations
from typing import TYPE_CHECKING
from typing import Any


from fenix.base.exchange import Exchange

from fenix.base.constants import Side
from fenix.base.constants import OrderType
from fenix.base.constants import ExchangeCode
from fenix.base.constants import Product
from fenix.base.constants import Validity
from fenix.base.constants import Variety
from fenix.base.constants import Status
from fenix.base.constants import Order
from fenix.base.constants import Profile
from fenix.base.constants import Position
from fenix.base.constants import Root
from fenix.base.constants import WeeklyExpiry
from fenix.base.constants import UniqueID


from fenix.base.errors import InputError
from fenix.base.errors import ResponseError
from fenix.base.errors import TokenDownloadError

import time
from datetime import datetime

if TYPE_CHECKING:
    from requests.models import Response



class paper(Exchange):
    """
    Paper fenix Broker Class.

    Returns:
        fenix.paper: fenix Paper Broker Object.
    """


    # Market Data Dictonaries

    indices = {}
    eq_tokens = {}
    nfo_tokens = {}
    token_params = ["user_id", "password", "birth_year", "totpstr", "api_key"]
    orderbook = []
    positions = {}
    id = 'paper'
    _session = Exchange._create_session()


    # Base URLs

    base_urls = {
        "market_data": "https://v2api.aliceblueonline.com/restpy/contract_master",
    }

    # Request Parameters Dictionaries

    req_exchange = {
        ExchangeCode.NSE: "NSE",
        ExchangeCode.NFO: "NFO",
        ExchangeCode.BSE: "BSE",
        ExchangeCode.BFO: "BFO",
        ExchangeCode.BCD: "BCD",
        ExchangeCode.MCX: "MCX",
        ExchangeCode.CDS: "CDS",
    }

    req_order_type = {
        OrderType.MARKET: "MKT",
        OrderType.LIMIT: "L",
        OrderType.SL: "SL",
        OrderType.SLM: "SL-M"
    }

    req_product = {
        Product.MIS: "MIS",
        Product.NRML: "NRML",
        Product.CNC: "CNC",
        Product.CO: "CO",
        Product.BO: "BO"
    }

    req_side = {
        Side.BUY: "BUY",
        Side.SELL: "SELL",
    }

    req_validity = {
        Validity.DAY: "DAY",
        Validity.IOC: "IOC",
    }

    req_variety = {
        Variety.REGULAR: "REGULAR",
        Variety.STOPLOSS: "REGULAR",
        Variety.BO: "BO",
        Variety.AMO: "AMO",
    }



    # response_dicts

    rms_limits_dict = {
        "symbol": "ALL",
        "cncMarginUsed": "0.00",
        "spanmargin": "0.00",
        "branchAdhoc": "0.000000",
        "adhocMargin": "0.00",
        "payoutamount": "0.00",
        "cdsSpreadBenefit": "0",
        "adhocscripmargin": "0.00",
        "exposuremargin": "0.00",
        "scripbasketmargin": "0.00",
        "credits": "0.00",
        "segment": "ALL",
        "net": "0.00",
        "turnover": "0.00",
        "grossexposurevalue": "0.00",
        "mfssAmountUsed": "0.00",
        "realizedMtomPrsnt": "0.00",
        "product": "ALL",
        "stat": "Ok",
        "cncSellCrditPrsnt": "0",
        "debits": "0.00",
        "varmargin": "0.00",
        "multiplier": "1.00",
        "elm": "0.00",
        "mfamount": "0.00",
        "cashmarginavailable": "0.00",
        "brokeragePrsnt": "0.00",
        "cncRealizedMtomPrsnt": "0",
        "notionalCash": "0.000000",
        "directcollateralvalue": "0.00",
        "cncBrokeragePrsnt": "0.00",
        "valueindelivery": "0",
        "nfoSpreadBenefit": "0",
        "losslimit": "0",
        "subtotal": "0.00",
        "rmsPayInAmnt": "0.00",
        "unrealizedMtomPrsnt": "0.00",
        "coverOrderMarginPrsnt": "0.00",
        "exchange": "ALL",
        "category": "ABFSFREEDOM",
        "collateralvalue": "0.00",
        "rmsIpoAmnt": "0",
        "cncUnrealizedMtomPrsnt": "0.00",
        "premiumPrsnt": "0.00"
    }

    profile_dict = {
        "accountStatus": "Activated",
        "dpType": "NA",
        "accountId": "796894",
        "sBrokerName": "PAPER",
        "product": ["CNC", "NRML", "MIS", "CO", "BO"],
        "accountName": "KRATOS",
        "cellAddr": "8750407885",
        "emailAddr": "GODOFWAR@GMAIL.COM",
        "exchEnabled": "bse_cm|bse_fo|bcs_fo|cde_fo|mcx_fo|nse_cm|nse_fo|"
    }


    # NFO Script Fetch


    @classmethod
    def create_eq_tokens(cls) -> dict:
        """
        Gives Indices Info for F&O Segment.
        Stores them in the paper.indices Dictionary.

        Returns:
            dict: Unified fenix indices format.
        """
        req_bse = cls.fetch(method="GET", url=cls.base_urls["market_data"], params={'exch': ExchangeCode.BSE})
        resp_bse = cls._json_parser(req_bse)

        df_bse = cls.data_frame(resp_bse[ExchangeCode.BSE])

        df_bse = df_bse[(df_bse['instrument_type'] == "E")]
        df_bse = df_bse[["trading_symbol", 'token', "tick_size", "lot_size", "exch"]]
        df_bse['token'] = df_bse['token'].astype(int)
        df_bse.rename({"token": "Token", "trading_symbol": "Symbol",
                       "tick_size": 'TickSize', "lot_size": "LotSize", "exch": "Exchange"}, axis=1, inplace=True)

        df_bse.set_index(df_bse['Symbol'], inplace=True)

        df_bse['Token'] = df_bse['Token']  # .astype(int)
        df_bse['TickSize'] = df_bse['TickSize'].astype(float)
        df_bse['LotSize'] = df_bse['LotSize'].astype(int)


        req_nse = cls.fetch(method="GET", url=cls.base_urls["market_data"], params={'exch': ExchangeCode.NSE})
        resp_nse = cls._json_parser(req_nse)

        df_nse = cls.data_frame(resp_nse[ExchangeCode.NSE])

        df_nse = df_nse[(df_nse['group_name'] == "EQ")]
        df_nse = df_nse[["symbol", "trading_symbol", 'token', "lot_size", "tick_size", "exch"]]
        df_nse.rename({"symbol": "Index", "token": "Token", "trading_symbol": "Symbol",
                       "tick_size": 'TickSize', "lot_size": "LotSize", "exch": "Exchange"}, axis=1, inplace=True)

        df_nse.set_index(df_nse['Index'], inplace=True)
        df_nse.drop(columns="Index", inplace=True)


        df_nse['Token'] = df_nse['Token'].astype(int)
        df_nse['TickSize'] = df_nse['TickSize'].astype(float)
        df_nse['LotSize'] = df_nse['LotSize'].astype(int)



        cls.eq_tokens[ExchangeCode.NSE] = df_nse.to_dict(orient='index')
        cls.eq_tokens[ExchangeCode.BSE] = df_bse.to_dict(orient='index')

        return cls.eq_tokens

    @classmethod
    def create_indices(cls) -> dict:
        """
        Gives Indices Info for F&O Segment.
        Stores them in the paper.indices Dictionary.

        Returns:
            dict: Unified fenix indices format.
        """
        req = cls.fetch(method="GET", url=cls.base_urls["market_data"], params={'exch': ExchangeCode.NSE})
        resp = cls._json_parser(req)

        df = cls.data_frame(resp[ExchangeCode.NSE])

        df = df[df["exchange_segment"] == "nse_idx"][["symbol", "token"]]

        df.rename({"symbol": "Symbol", "token": "Token"}, axis=1, inplace=True)
        df.index = df['Symbol']

        indices = df.to_dict(orient='index')

        indices[Root.BNF] = indices["NIFTY BANK"]
        indices[Root.NF] = indices["NIFTY 50"]
        indices[Root.FNF] = indices["NIFTY FIN SERVICE"]
        indices[Root.MIDCPNF] = indices["NIFTY MIDCAP SELECT"]

        cls.indices = indices

        return indices

    @classmethod
    def create_nfo_tokens(cls) -> dict:
        """
        Creates BANKNIFTY & NIFTY Current, Next and Far Expiries;
        Stores them in the paper.nfo_tokens Dictionary.

        Raises:
            TokenDownloadError: Any Error Occured is raised through this Error Type.
        """
        try:
            req = cls.fetch(method="GET", url=cls.base_urls["market_data"], params={'exch': ExchangeCode.NFO})
            resp = cls._json_parser(req)

            df = cls.data_frame(resp[ExchangeCode.NFO])

            df = df[
                (
                    (df['symbol'] == 'BANKNIFTY') |
                    (df['symbol'] == 'NIFTY') |
                    (df['symbol'] == 'FINNIFTY') |
                    (df['symbol'] == 'MIDCPNIFTY')
                ) &
                (
                    (df['instrument_type'] == 'OPTIDX')
                )
            ]

            df.rename({"option_type": "Option", "token": "Token", "symbol": "Root", "exch": "Exchange",
                       "expiry_date": "Expiry", "trading_symbol": "Symbol",
                       "tick_size": 'TickSize', "lot_size": "LotSize", "strike_price": "StrikePrice"
                       }, axis=1, inplace=True)
            df = df[['Token', 'Symbol', 'Expiry', 'Option', 'StrikePrice',
                     'LotSize', 'Root', 'TickSize', "Exchange"
                     ]]

            df['StrikePrice'] = df['StrikePrice'].astype(int)
            df['Expiry'] = cls.pd_datetime(df['Expiry'], unit='ms').dt.date.astype(str)
            df["Token"] = df["Token"].astype(int)
            df["TickSize"] = df["TickSize"].astype(float)
            df["LotSize"] = df["LotSize"].astype(float)

            expiry_data = cls.jsonify_expiry(data_frame=df)
            cls.nfo_tokens = expiry_data

            return expiry_data

        except Exception as exc:
            raise TokenDownloadError({"Error": exc.args}) from exc


    # Headers & Json Parsers


    @classmethod
    def create_headers(cls,
                       params: dict = {},
                       ) -> dict[str, str]:
        """
        Generate Headers used to access Endpoints in Paper.

        Parameters:
            Params (dict) : A placeholder dictionary for header creation.

        Returns:
            dict[str, str]: Paper Headers.
        """

        headers = {
            "headers":
                {
                    "ID": "user_id",
                    "AccessToken": "access_token",
                    "Authorization": "Bearer user_id access_token",
                    "X-SAS-Version": "2.0",
                    "User-Agent": "fenix_Paper",
                    "Content-Type": "application/json",
                    "susertoken": "susertoken"
                }
        }


        return headers

    @classmethod
    def _json_parser(cls,
                     response: Response
                     ) -> dict[Any, Any] | list[dict[Any, Any]]:
        """
        Parses the Json Repsonse Obtained from Broker.

        Parameters:
            response (Response): Json Response Obtained from Broker.

        Raises:
            ResponseError: Raised if any error received from broker.

        Returns:
            dict: json response obtained from exchange.
        """
        json_response = cls.on_json_response(response)
        if isinstance(json_response, dict):
            stat = json_response.get('stat', None)

            if stat == 'Ok' or not stat:
                return json_response

            error = json_response.get('emsg', None)
            error = error if error else json_response.get('Emsg', None)
            raise ResponseError(cls.id + " " + error)

        if isinstance(json_response, list):
            stat = json_response[0].get('stat', )

            if stat == 'Ok' or not stat:
                return json_response

            error = json_response[0].get('emsg', None)
            error = error if error else json_response[0].get('Emsg', str(json_response))
            raise ResponseError(cls.id + " " + error)

    @classmethod
    def _order_data_creator(cls,
                            order: dict,
                            ) -> dict[Any, Any]:
        """
        Parse Orderbook Order Json Response.

        Parameters:
            order (dict): Orderbook Order Json Response from Broker.

        Returns:
            dict: Unified fenix Order Response.
        """
        parsed_order = {
            Order.ID: str(time.time_ns()),
            Order.USERID: order[Order.USERID],
            Order.TIMESTAMP: datetime.now().replace(microsecond=0),
            Order.SYMBOL: order[Order.SYMBOL],
            Order.TOKEN: order[Order.TOKEN],
            Order.SIDE: order[Order.SIDE],
            Order.TYPE: order[Order.TYPE],
            Order.AVGPRICE: order[Order.PRICE],
            Order.PRICE: order[Order.PRICE],
            Order.TRIGGERPRICE: order[Order.TRIGGERPRICE],
            Order.TARGETPRICE: 0.0,
            Order.STOPLOSSPRICE: 0.0,
            Order.TRAILINGSTOPLOSS: 0.0,
            Order.QUANTITY: order[Order.QUANTITY],
            Order.FILLEDQTY: order[Order.QUANTITY],
            Order.REMAININGQTY: 0,
            Order.CANCELLEDQTY: 0,
            Order.STATUS: Status.PENDING,
            Order.REJECTREASON: "",
            Order.DISCLOSEDQUANTITY: 0,
            Order.PRODUCT: order[Order.PRODUCT],
            Order.EXCHANGE: order[Order.EXCHANGE],
            Order.SEGMENT: order[Order.EXCHANGE],
            Order.VALIDITY: order[Order.VALIDITY],
            Order.VARIETY: "",
            Order.INFO: {},
        }

        cls.orderbook.append(parsed_order)
        return parsed_order

    @classmethod
    def _profile_paper(cls):
        """
        Parse User Profile Json Response.

        Parameters:
            profile (dict): User Profile Json Response from Broker.

        Returns:
            dict: Unified fenix Profile Response.
        """
        exchanges_enabled = [ExchangeCode.NSE, ExchangeCode.BSE, ExchangeCode.NFO,
                             ExchangeCode.BFO, ExchangeCode.MCX, ExchangeCode.CDS,
                             ExchangeCode.BFO, ExchangeCode.BCD, ExchangeCode.NCO,
                             ExchangeCode.BCO
                             ]

        parsed_profile = {
            Profile.CLIENTID: "796894",
            Profile.NAME: "KRATOS",
            Profile.EMAILID: "GODOFWAR@GMAIL.COM",
            Profile.MOBILENO: "7054251468",
            Profile.PAN: "KRATOSPANNO",
            Profile.ADDRESS: "GREEK MYTHOS",
            Profile.BANKNAME: "OMEGA BANK",
            Profile.BANKBRANCHNAME: "NO BRANCH",
            Profile.BANKACCNO: "5354983205",
            Profile.EXHCNAGESENABLED: exchanges_enabled,
            Profile.ENABLED: True,
            Profile.INFO: cls.profile_dict
        }

        return parsed_profile


    # Order Functions


    @classmethod
    def create_eq_nfo_order(cls,
                            quantity: int,
                            side: str,
                            headers: dict,
                            token_dict: dict,
                            price: float = 0.0,
                            trigger: float = 0.0,
                            product: str = Product.MIS,
                            validity: str = Validity.DAY,
                            variety: str = Variety.REGULAR,
                            unique_id: str = UniqueID.DEFORDER
                            ) -> dict[Any, Any]:
        """
        Place an Order in F&O and Equity Segment.

        Parameters:
            quantity (int): Order quantity.
            side (str): Order Side: "BUY", "SELL".
            headers (dict): headers to send order request with.
            token_dict (dict): a dictionary with details of the Ticker. Obtianed from eq_tokens or nfo_tokens.
            price (float): price of the order. Defaults to 0.0.
            trigger (float): trigger price of the order. Defaults to 0.0.
            product (str, optional): Order product. Defaults to Product.MIS.
            validity (str, optional): Order validity Defaults to Validity.DAY.
            variety (str, optional): Order variety Defaults to Variety.REGULAR.
            unique_id (str, optional): Unique user orderid. Defaults to UniqueID.DEFORDER.

        Returns:
            dict: fenix Unified Order Response.
        """
        if not price and trigger:
            order_type = OrderType.SLM
        elif not price:
            order_type = OrderType.MARKET
        elif not trigger:
            order_type = OrderType.LIMIT
        else:
            order_type = OrderType.SL

        token = token_dict["Token"]
        exchange = token_dict["Exchange"]
        symbol = token_dict["Symbol"]

        order = {
            Order.USERID: unique_id,
            Order.SYMBOL: symbol,
            Order.TOKEN: token,
            Order.SIDE: cls._key_mapper(cls.req_side, side, "side"),
            Order.TYPE: order_type,
            Order.PRICE: price,
            Order.TRIGGERPRICE: 0,
            Order.QUANTITY: quantity,
            Order.PRODUCT: cls._key_mapper(cls.req_product, product, "product"),
            Order.EXCHANGE: exchange,
            Order.SEGMENT: exchange,
            Order.VALIDITY: cls._key_mapper(cls.req_validity, validity, "validity"),
            Order.VARIETY: "",
            Order.INFO: {},
        }

        order = cls._order_data_creator(order)
        return order


    # Order Details, OrderBook & TradeBook


    @classmethod
    def fetch_orderbook(cls,
                        headers: dict
                        ) -> list[dict]:
        """
        Fetch Orderbook Details.

        Parameters:
            headers (dict): headers to send fetch_orders request with.

        Returns:
            list[dict]: List of dicitonaries of orders using fenix Unified Order Response.
        """
        return cls.orderbook

    @classmethod
    def fetch_tradebook(cls,
                        headers: dict
                        ) -> list[dict]:
        """
        Fetch Tradebook Details.

        Parameters:
            headers (dict): headers to send fetch_orders request with.

        Returns:
            list[dict]: List of dicitonaries of orders using fenix Unified Order Response.
        """
        return cls.orderbook

    @classmethod
    def fetch_orders(cls,
                     headers: dict
                     ) -> list[dict]:
        """
        Fetch OrderBook Details which is unified across all brokers.
        Use This if you want Avg price, etc. values which sometimes unavailable
        thorugh fetch_orderbook.

        Paramters:
            order_id (str): id of the order.

        Raises:
            InputError: If order does not exist.

        Returns:
            dict: fenix Unified Order Response.
        """
        return cls.orderbook

    @classmethod
    def fetch_order(cls,
                    order_id: str,
                    headers: dict
                    ) -> dict[Any, Any]:
        """
        Fetch Order Details.

        Paramters:
            order_id (str): id of the order.

        Raises:
            InputError: If order does not exist.

        Returns:
            dict: fenix Unified Order Response.
        """
        order_id = str(order_id)

        for order in cls.orderbook:
            if order[Order.ID] == order_id:
                return order

        raise InputError({"This order_id does not exist."})

    @classmethod
    def fetch_orderhistory(cls,
                           order_id: str,
                           headers: dict
                           ) -> list[dict]:
        """
        Fetch History of an order.

        Paramters:
            order_id (str): id of the order.
            headers (dict): headers to send orderhistory request with.

        Returns:
            list: A list of dicitonaries containing order history using fenix Unified Order Response.
        """
        return [cls.fetch_order(order_id, headers)]


    # Order Modification & Sq Off


    @classmethod
    def modify_order(cls,
                     order_id: str,
                     headers: dict,
                     price: float | None = None,
                     trigger: float | None = None,
                     quantity: int | None = None,
                     order_type: str | None = None,
                     validity: str | None = None,
                     ) -> dict[Any, Any]:
        """
        Modify an open order.

        Parameters:
            order_id (str): id of the order to modify.
            headers (dict): headers to send modify_order request with.
            price (float | None, optional): price of t.he order. Defaults to None.
            trigger (float | None, optional): trigger price of the order. Defaults to None.
            quantity (int | None, optional): order quantity. Defaults to None.
            order_type (str | None, optional): Type of Order. defaults to None.
            validity (str | None, optional): Order validity Defaults to None.

        Returns:
            dict: fenix Unified Order Response.
        """
        order_id = str(order_id)

        for index in range(len(cls.orderbook)):
            if cls.orderbook[index][Order.ID] == order_id:
                order = cls.orderbook[index]
                if order[Order.STATUS] == Status.PENDING:
                    order[Order.PRICE] = price or order[Order.PRICE]
                    order[Order.TRIGGERPRICE] = trigger or order[Order.TRIGGERPRICE]
                    order[Order.QUANTITY] = quantity or order[Order.QUANTITY]
                    order[Order.VALIDITY] = cls._key_mapper(cls.req_validity, validity, 'validity')
                    order[Order.TYPE] = cls._key_mapper(cls.req_order_type, order_type, 'order_type') or order[Order.TYPE]

                return order

        raise InputError({"This order_id does not exist."})

    @classmethod
    def cancel_order(cls,
                     order_id: str,
                     headers: dict
                     ) -> dict[Any, Any]:
        """
        Cancel an open order.

        Parameters:
            order_id (str): id of the order.
            headers (dict): headers to send cancel_order request with.

        Returns:
            dict: fenix Unified Order Response.
        """
        order_id = str(order_id)

        for index in range(len(cls.orderbook)):
            if cls.orderbook[index][Order.ID] == order_id:
                order = cls.orderbook[index]
                if order[Order.STATUS] == Status.PENDING:
                    order[Order.STATUS] = Status.CANCELLED

                return order


        raise InputError({"This order_id does not exist."})


    # Positions, Account Limits & Profile


    @classmethod
    def _position_calc(cls,
                       position,
                       order,
                       key
                       ):

        qkey = f"{key}qty"
        pkey = f"{key}price"

        position[Position.INFO][qkey].append(order[Order.QUANTITY])
        position[Position.INFO][pkey].append(order[Order.AVGPRICE])

        qty = sum(position[Position.INFO][qkey])
        price = 0

        for order_qty, order_price in zip(position[Position.INFO][qkey], position[Position.INFO][pkey]):
            price += (order_qty * order_price)

        avgprice = price / qty

        return qty, avgprice, price

    @classmethod
    def fetch_day_positions(cls,
                            headers: dict,
                            ) -> dict[Any, Any]:
        """
        Fetch the Day's Account Positions.

        Args:
            headers (dict): headers to send rms_limits request with.

        Returns:
            dict[Any, Any]: fenix Unified Position Response.
        """
        return cls.fetch_positions(headers)

    @classmethod
    def fetch_net_positions(cls,
                            headers: dict,
                            ) -> dict[Any, Any]:
        """
        Fetch Total Account Positions.

        Args:
            headers (dict): headers to send rms_limits request with.

        Returns:
            dict[Any, Any]: fenix Unified Position Response.
        """
        return cls.fetch_positions(headers)

    @classmethod
    def fetch_positions(cls,
                        headers: dict,
                        ) -> dict[Any, Any]:
        """
        Fetch Day & Net Account Positions.

        Args:
            headers (dict): headers to send rms_limits request with.

        Returns:
            dict[Any, Any]: fenix Unified Position Response.
        """
        for order in cls.orderbook:

            if order[Order.STATUS] == Status.FILLED:
                if order[Order.SYMBOL] in cls.positions:
                    position = cls.positions[order[Order.SYMBOL]]

                    if order[Order.SIDE] == Side.BUY:
                        position[Position.BUYQTY], position[Position.BUYPRICE], position[Position.INFO]["b_avgprice"] = cls._position_calc(position, order, "b")
                    else:
                        position[Position.SELLQTY], position[Position.SELLPRICE], position[Position.INFO]["s_avgprice"] = cls._position_calc(position, order, "s")

                    tprice = position[Position.INFO]["b_avgprice"] + position[Position.INFO]["s_avgprice"]
                    tqty = position[Position.BUYQTY] + position[Position.SELLQTY]

                    position[Position.NETQTY] = abs(position[Position.BUYQTY] - position[Position.SELLQTY])
                    position[Order.AVGPRICE] = tprice / tqty

                else:
                    position = {
                        Position.SYMBOL: order[Order.SYMBOL],
                        Position.TOKEN: order[Order.TOKEN],
                        Position.NETQTY: order[Order.QUANTITY],
                        Position.AVGPRICE: order[Order.AVGPRICE],
                        Position.MTM: None,
                        Position.PNL: None,
                        Position.BUYQTY: 0,
                        Position.BUYPRICE: 0.0,
                        Position.SELLQTY: 0,
                        Position.SELLPRICE: 0.0,
                        Position.LTP: None,
                        Position.PRODUCT: order[Order.PRODUCT],
                        Position.EXCHANGE: order[Order.EXCHANGE],
                        Position.INFO: {"bqty": [], "bprice": [], "b_avgprice": 0,
                                        "sqty": [], "sprice": [], "s_avgprice": 0}
                    }


                    if order[Order.SIDE] == Side.BUY:
                        position[Position.BUYQTY] = order[Order.QUANTITY]
                        position[Position.BUYPRICE] = order[Order.AVGPRICE]
                        position[Position.INFO]["bqty"].append(order[Order.QUANTITY])
                        position[Position.INFO]["bprice"].append(order[Order.AVGPRICE])
                    else:
                        position[Position.SELLQTY] = order[Order.QUANTITY]
                        position[Position.SELLPRICE] = order[Order.AVGPRICE]
                        position[Position.INFO]["sqty"].append(order[Order.QUANTITY])
                        position[Position.INFO]["sprice"].append(order[Order.AVGPRICE])


                    cls.positions[order[Order.SYMBOL]] = position


        return list(cls.positions.values())


    @classmethod
    def fetch_holdings(cls,
                       headers: dict,
                       ) -> dict[Any, Any]:
        """
        Fetch Account Holdings.

        Args:
            headers (dict): headers to send rms_limits request with.

        Returns:
            dict[Any, Any]: fenix Unified Positions Response.
        """
        return cls.fetch_positions(headers)

    @classmethod
    def rms_limits(cls,
                   headers: dict
                   ) -> dict[Any, Any]:
        """
        Fetch Risk Management System Limits.

        Parameters:
            headers (dict): headers to send rms_limits request with.

        Returns:
            dict: fenix Unified RMS Limits Response.
        """
        return cls.rms_limits_dict

    @classmethod
    def profile(cls,
                headers: dict
                ) -> dict[Any, Any]:
        """
        Fetch Profile Limits of the User.

        Parameters:
            headers (dict): headers to send profile request with.

        Returns:
            dict: fenix Unified Profile Response.
        """
        return cls._profile_paper()



from __future__ import annotations
from datetime import datetime
from datetime import timedelta
from json import dumps
from json import loads
from ssl import SSLError
from typing import Any
from pandas import DataFrame
from pandas import read_csv
from pandas import read_json
from pandas import to_datetime
from pandas.tseries.offsets import DateOffset
from pandas import options
from pyotp import TOTP

from re import compile


from requests.sessions import session as req_session
from requests.models import Response
from requests.exceptions import HTTPError
from requests.exceptions import Timeout
from requests.exceptions import TooManyRedirects
from requests.exceptions import RequestException
from requests.exceptions import ConnectionError as requestsConnectionError

from fenix.base.constants import Root
from fenix.base.constants import WeeklyExpiry

from fenix.base.errors import InputError
from fenix.base.errors import ResponseError
from fenix.base.errors import RequestTimeout
from fenix.base.errors import NetworkError
from fenix.base.errors import BrokerError

options.mode.chained_assignment = None


__all__ = [
    'Exchange',
]


class Exchange:

    """ Base Class Common to All Brokers """

    id = ''
    indices = {}
    eq_tokens = {}
    nfo_tokens = {}
    _session = None

    def __repr__(self):
        return f"fenix.{self.id}()"

    @staticmethod
    def _create_session():
        """
        Creates A request Session.

        Returns:
            requests.sessions.session: returns a reqeusts.sessions.session Object
        """
        return req_session()

    @classmethod
    def fetch(cls,
              method: str,
              url: str,
              headers: dict[Any, Any] | None = None,
              data: dict[Any, Any] | None = None,
              json: dict[Any, Any] | None = None,
              params: dict[Any, Any] | None = None,
              auth: tuple[str, str] | None = None,
              timeout: int = 10,
              ) -> Response:
        """
        A Wrapper for Python Requests module,
        sending requests over a session which persists the cookies over the entire session.

        Parameters:
            method (str): Request Method: 'GET', 'POST', 'PUT', 'DELETE', 'GET', etc.
            url (str): URL of the Request
            headers (dict[Any, Any] | None, optional): Request Headers. Defaults to None.
            data (dict[Any, Any] | None, optional): Dictionary, list of tuples, bytes, or file-like object to send in the body of the Request. Defaults to None.
            json (dict[Any, Any] | None, optional): A JSON serializable Python object to send in the body of the Request. Defaults to None.
            params (dict[Any, Any] | None, optional): Dictionary, list of tuples or bytes to send in the query string for the Request. Defaults to None.
            auth (tuple[str, str] | None, optional): Auth tuple to enable Basic/Digest/Custom HTTP Auth. Defaults to None.
            timeout (int, optional): How many seconds to wait for the server to send data before giving up. Defaults to 10.

        Raises:
            RequestTimeout: If the Request Times Out
            NetworkError: If Network Unavailable.
            BrokerError: If Error on Behalf of the Broker or Some Error on behalf of the User sending the Request.

        Returns:
            Response: Response Object
        """

        try:
            response = cls._session.request(method=method,
                                            url=url,
                                            headers=headers,
                                            data=data,
                                            json=json,
                                            params=params,
                                            auth=auth,
                                            timeout=timeout,
                                            )

            response.raise_for_status()

            return response

        except Timeout as exc:
            details = ' '.join([cls.id, method, url])
            raise RequestTimeout(details) from exc

        except TooManyRedirects as exc:
            details = ' '.join([cls.id, method, url])
            raise BrokerError(details) from exc

        except SSLError as exc:
            details = ' '.join([cls.id, method, url])
            raise BrokerError(details) from exc

        except HTTPError as exc:
            details = ' '.join([cls.id, method, str(response.status_code), url, response.text])
            raise BrokerError(details) from exc

        except requestsConnectionError as exc:
            error_string = str(exc)
            details = ' '.join([cls.id, method, url, error_string])
            if 'Read timed out' in error_string:
                raise RequestTimeout(details) from exc
            raise NetworkError(details) from exc

        except ConnectionResetError as exc:
            details = ' '.join([cls.id, method, url])
            raise NetworkError(details) from exc

        except RequestException as exc:  # base exception class
            error_string = str(exc)
            details = ' '.join([cls.id, method, url])
            if any(x in error_string for x in ['ECONNRESET', 'Connection aborted.', 'Connection broken:']):
                raise NetworkError(exc) from exc
            raise BrokerError(exc) from exc

    @staticmethod
    def _json_parser(response: Response) -> dict[Any, Any] | list[dict[Any, Any]]:
        """
        Get json object from a request Response.

        Parameters:
            response (Response): Response Object

        Returns:
            dict: Json Object received from Response.
        """
        try:
            response = loads(response.text.strip())
            # print(response)
            return response
        except Exception as exc:
            raise ResponseError({"Status": response.status_code,
                                 "Error": response.text,
                                 "URL": response.url,
                                 "Reason": response.reason
                                 }
                                ) from exc

    @staticmethod
    def json_dumps(json_data: dict) -> str:
        return dumps(json_data)

    @staticmethod
    def on_json_response(response: Response) -> dict[Any, Any]:
        """
        Get json object from a request Response.

        Parameters:
            response (Response): Response Object

        Returns:
            dict: Json Object received from Response.
        """
        # print(response.text)
        return loads(response.text.strip())


    @staticmethod
    def _eq_mapper(dictionary: dict, key: str) -> str:
        """
        A Simple Function to help the User if they input a wrong Symbol in the eq_tokens dictionary,
        also tells the User the possible Symbols for the Segment.

        Parameters:
            dictionary (dict): Dicitonary
            key (str): Dictionary Key to Check Against, should be capital Letters.

        Raises:
            KeyError: If Key Does not exist in the Dicitonary.

        Returns:
            str: The Value of the Key in the Dicitonary.
        """
        key = key.upper()
        if key in dictionary:
            return dictionary[key]

        r = compile(r"[A-Z]*<str>[A-Z]*$".replace("<str>", key))
        possible_values = list(filter(r.findall, dictionary))

        raise KeyError(f"Invalid Symbol!: {key}, Possible Values: {possible_values}")


    @staticmethod
    def _key_mapper(dictionary: dict, key: str, name: str) -> str:
        """
        A Simple Function to help the User if they input a wrong Key in the Dictionary,
        also tells the User the possible Keys for the Dicitonary.

        Parameters:
            dictionary (dict): Dicitonary
            key (str): Dictionary Key to Check Against.
            name (str): The right Key to the Dicitonary.

        Raises:
            KeyError: If Key Does not exist in the Dicitonary.

        Returns:
            str: The Value of the Key in the Dicitonary.
        """

        if key in dictionary:
            return dictionary[key]

        raise KeyError(f"Invalid {name}!: {key}, Possible Values: {list(dictionary.keys())}")

    @staticmethod
    def totp_creator(totpbase: str) -> str:
        """
        Get TOTP from the character string.

        Parameters:
            totpbase (str): String used to Generate TOTP.

        Returns:
            str: Six-Digit TOTP
        """
        while True:
            totpobj = TOTP(totpbase)
            totp = totpobj.now()

            if totpobj.verify(totp):
                return totp


    @staticmethod
    def data_reader(link: str,
                    filetype: str,
                    dtype: dict | None = None,
                    sep: str = ",",
                    col_names: list = [],
                    ) -> DataFrame:
        """
        Pandas.read_csv & Pandas.read_json Functions Wrapper

        Parameters:
            link (str): URL to get the Data From.
            filetype (str): 'json' | 'csv'
            dtype (dict | None, optional): A Dicitonary with Column-Names as the Keys and their Datatypes as the values. Defaults to None.
            sep (str, optional): Needed with filetype as 'csv', to input the data seperator. Defaults to ','.

        Raises:
            InputError: If Wrong filetype Given as Input

        Returns:
            DataFrame: Pandas DataFrame
        """
        if filetype == "json":
            return read_json(link)

        if filetype == "csv":
            if col_names:
                return read_csv(link,
                                dtype=dtype,
                                sep=sep,
                                names=col_names,
                                )

            return read_csv(link,
                            dtype=dtype,
                            sep=sep
                            )


        raise InputError(f"Wrong Filetype: {filetype}, the possible values are: 'json', 'csv'")

    @staticmethod
    def data_frame(data: list) -> DataFrame:
        """
        Pandas.DataFrame Function Wrapper

        Parameters:
            data (list): List of Data to make the DataFrame out of.

        Returns:
            DataFrame: Pandas DataFrame
        """
        return DataFrame(data)


    @staticmethod
    def pd_datetime(datetime_obj: str,
                    unit: str = "ns"
                    ) -> datetime:
        """
        Pandas.to_datetime Function Wrapper

        Parameters:
            datetime_obj (str): Datetime String to convert to datetime object.
            unit (str, optional): The unit of the datetime_obj (D,s,ms,us,ns) denote the unit. Defaults to "ns".

        Returns:
            datetime: Timestamp object
        """
        return to_datetime(arg=datetime_obj,
                           unit=unit
                           )

    @staticmethod
    def datetime_strp(datetime_obj: str,
                      dtformat: str
                      ) -> datetime:
        """
        Python datetime.datetime.strptime Function Wrapper

        Parameters:
            datetime_obj (str): Datetime String to convert to datetime object.
            dtformat (str): corresponding datetime format string.

        Returns:
            datetime: datetime.datetime object
        """
        return datetime.strptime(datetime_obj,
                                 dtformat
                                 )

    @staticmethod
    def from_timestamp(datetime_obj: int) -> datetime:
        """
        Convert Epoch Time to datetime.datetime object

        Parameters:
            datetime_obj (int): Epoch datetime

        Returns:
            datetime: datetime.datetime object
        """
        return datetime.fromtimestamp(datetime_obj)

    @staticmethod
    def current_datetime() -> datetime:
        """
        Get Current System Datetime

        Returns:
            datetime: datetime.datetime object
        """
        return datetime.now()

    @staticmethod
    def time_delta(datetime_object: datetime,
                   delta: int,
                   dtformat: str,
                   default='sub'
                   ) -> str:
        """
        Add Days to a datetime.datetime object

        Parameters:
            datetime_object (datetime): datetime object
            delta (int): No. of Days to add or subtract from datetime_obj
            dtformat (str): corresponding datetime format string.
            default (str, optional): Whether to add or subtract a Day from datetime_obj ('add' | 'sub'). Defaults to 'sub'.

        Raises:
            InputError: If Wrong Value Given for default Parameter.

        Returns:
            str: A datetime string.
        """
        if default == "sub":
            return (datetime_object - timedelta(days=delta)).strftime(dtformat)
        if default == "add":
            return (datetime_object + timedelta(days=delta)).strftime(dtformat)

        raise InputError(f"Wrong default: {default}, the possible values are 'sub', 'add'")

    def pd_dateoffset(*args, **kwargs) -> DateOffset:
        """
        Datetime Increment from Pandas.

        Returns:
            DateOffset: Pandas DateOffset.
        """
        return DateOffset(*args, **kwargs)




    @staticmethod
    def jsonify_expiry(data_frame: DataFrame) -> dict[Any, Any]:
        """
        Creates a fenix Unified Dicitonary for BankNifty & Nifty Options,
        in the following Format:

            Global[expiry][root][option][strikeprice]

            expiry: 'CURRENT' | 'NEXT' | 'FAR' | 'Expiries'(A List of All Expiries).
            root: 'BANKNIFTY' | 'NIFTY' | 'FINNIFTY | MIDCPNIFTY.
            option: 'CE' | 'PE'.
            strikeprice: Integer Values for Strike Price.

            {
                'CURRENT': {
                    'BANKNIFTY': {
                        'CE': {
                            34500: {
                                'Token': 43048,
                                'Symbol': 'BANKNIFTY2331634500CE',
                                'Expiry': '2023-03-16',
                                'Option': 'CE',
                                'StrikePrice': 34500,
                                'LotSize': 25,
                                'Root': 'BANKNIFTY',
                                'TickSize': 0.05,
                                'ExpiryName': 'CURRENT'
                                },
                    ...

                'Expiries': [
                     '2023-03-16', '2023-03-23', '2023-03-29', '2023-04-06',
                     '2023-04-13', '2023-04-27', '2023-05-25', '2023-06-29',
                     '2023-09-28', '2023-12-28', '2024-06-28', '2024-12-27',
                     '2025-06-27', '2025-12-25', '2026-06-25', '2026-12-31',
                     '2027-06-24', '2027-12-30']

            }

        Parameters:
            data_frame (DataFrame): DataFrame to Convert to fenix Unified Expiry Dictionary

        Returns:
            dict[Any, Any]: Dictioanry With 3 Most Recent Expiries for Both BankNifty & Nifty.
        """

        expiry_data = {
            WeeklyExpiry.CURRENT: {Root.BNF: {}, Root.NF: {}, Root.FNF: {}, Root.MIDCPNF: {}},
            WeeklyExpiry.NEXT: {Root.BNF: {}, Root.NF: {}, Root.FNF: {}, Root.MIDCPNF: {}},
            WeeklyExpiry.FAR: {Root.BNF: {}, Root.NF: {}, Root.FNF: {}, Root.MIDCPNF: {}},
            WeeklyExpiry.EXPIRY: {Root.BNF: [], Root.NF: [], Root.FNF: [], Root.MIDCPNF: []},
            WeeklyExpiry.LOTSIZE: {Root.BNF: [], Root.NF: [], Root.FNF: [], Root.MIDCPNF: []},
        }

        data_frame = data_frame.sort_values(by=['Expiry'])

        for root in [Root.BNF, Root.NF, Root.FNF, Root.MIDCPNF]:


            expiries = data_frame[data_frame['Root'] == root]['Expiry'].unique()
            expiries = expiries[expiries >= str(datetime.now().date())]
            lotsize = data_frame[data_frame['Root'] == root]['LotSize'].unique()[0]

            dfex1 = data_frame[data_frame['Expiry'] == expiries[0]]
            dfex2 = data_frame[data_frame['Expiry'] == expiries[1]]
            dfex3 = data_frame[data_frame['Expiry'] == expiries[2]]

            dfexs = [("CURRENT", dfex1), ("NEXT", dfex2), ("FAR", dfex3)]

            for expiry_name, dfex in dfexs:

                dfex['ExpiryName'] = expiry_name

                global_dict = {'CE': {}, "PE": {}}

                for j, i in dfex.groupby(['Option', 'StrikePrice']):
                    global_dict[j[0]][j[1]] = i.to_dict('records')[0]
                    expiry_data[expiry_name][root] = global_dict

                expiry_data['Expiry'][root] = list(expiries)
                expiry_data['LotSize'][root] = lotsize

        return expiry_data



import setuptools
import codecs
import os.path

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name="fenix",
    version=get_version("fenix/__init__.py"),
    keywords="fenix, aliceblue, angelone, choice, finvasia, fivepaisa, fyers, iifl, kotak, kotakneo, kunjee, mastertrust, motilaloswal, paper, symphony, upstox, vpc, zerodha, finance, broker, trader, XTS, kite, algorithmic, algotrading, api, arbitrage, real-time, realtime, backtest, backtesting, etc, framework, invest, investing, investor, library, market, market data, ohlcv, order, orderbook, order book, strategy, ticker, tickers, toolkit, trade, trader, trading, volume, websocket, websockets, web socket, web sockets, ws",
    author="Hardeep Singh",
    author_email="hardeep.hd13@gmail.com",
    description="A Python library for trading in the Indian Finance Sector with support for multiple broker APIs.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license="GPLv3",
    url="https://fenix.hardeep.tech",
    packages=setuptools.find_packages(),
    python_requires=">=3.8",
    install_requires=[
        "cryptography>=42.0.5",
        "numpy>=1.24.4",
        "pandas>=2.0.3",
        "pycryptodome>=3.20.0",
        "PyJWT>=2.8.0",
        "pyotp>=2.9.0",
        "requests>=2.31.0",
        "requests-oauthlib>=1.4.0",
        "selenium>=4.18.1",
        "urllib3>=2.2.1",
        "webdriver-manager>=4.0.1",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Information Technology",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Office/Business :: Financial :: Investment",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
        "Environment :: Console",
    ],
    project_urls={
        "Documentation": "https://github.com/TheHardeep/fenix/fenix/wiki",
        "Source Code": "https://github.com/TheHardeep/fenix.git",
    },
)



# Hardeep's Unified Library For Indian Brokers
# This module follows the GPL3 Open Source License


from fenix.base import exchange
from fenix.base import errors
from fenix.base import constants


__all__ = exchange.__all__ + errors.__all__ + constants.__all__   # noqa: F405
